//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Library General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include "CreateScenario.h"
#include "army.h"
#include "GameMap.h"
#include "counter.h"
#include "real_player.h"
#include "ai_fast.h"
#include "ai_dummy.h"
#include "File.h"
#include "MapCreationDialog.h"
#include "QuestsManager.h"
#include "Configuration.h"


#define debug(x) {cerr<<__FILE__<<": "<<__LINE__<<": "<<x<<endl;}
//#define debug(x)



CreateScenario::CreateScenario(PG_Widget* uncle)
    :d_generator(0), d_uncle(uncle)
{
    debug("CreateScenario::CreateScenario")
   
   //default value
    setSize(100);
    
    //make sure that objects are deleted
    GameMap::deleteInstance();
    Playerlist::deleteInstance();
    Templelist::deleteInstance();
    Ruinlist::deleteInstance();
    Citylist::deleteInstance();

    QuestsManager::deleteInstance();

    //this has to exist and be loaded
    Armysetlist::getInstance()->loadArmyset("default");
    
    fl_counter = new FL_Counter();
}

CreateScenario::~CreateScenario()
{
    debug("CreateScenario::~CreateScenario")

    if (d_generator)
        delete d_generator;

    // see GameScenario
    GameMap::deleteInstance();
	Armysetlist::deleteInstance();
    Playerlist::deleteInstance();
    Templelist::deleteInstance();
    Ruinlist::deleteInstance();
    Citylist::deleteInstance();

    QuestsManager::deleteInstance();
    
    delete fl_counter;
}

void CreateScenario::setMaptype(MapType type)
{
    debug("CreateScenario::setMaptype")

    //this function currently does almost nothing. It's purpose is to initialize
    //the right MapGenerator (e.g. islands, mountain chains, normal etc.).
    //Currently we have only a normal d_generator, so it is initialized here

    if (d_generator)
        delete d_generator;

    d_generator = new MapGenerator();
}

void CreateScenario::setPercentages(int pgrass, int pwater, int pforest,
                                    int pswamp, int phills, int pmountains)
{
    debug("CreateScenario::setPercentages")

    if (!d_generator)
        setMaptype(NORMAL);

    //handle input with !=100% sum
    int sum = pgrass + pwater + pforest + pswamp +phills + pmountains;

    if (sum != 100)
    {
        double factor = 100 / static_cast<double>(sum);  

        pwater = static_cast<int>(pwater * factor);
        pforest = static_cast<int>(pforest * factor);
        pswamp = static_cast<int>(pswamp * factor);
        phills = static_cast<int>(phills * factor);
        pmountains = static_cast<int>(pmountains * factor);
    }

    //the multiplication doesn't round up, so the figures should be OK now, the
    //missing percentage is implicitely added to the grass part.
    d_generator->setPercentages(pwater, pforest, pswamp, phills, pmountains);
}

void CreateScenario::setMapTiles(std::string tilesname)
{
    debug("CreateScenario::setMapTiles")
    d_tilesname = tilesname;
}

void CreateScenario::setNoCities(int nocities)
{
    debug("CreateScenario::setNoCities")

    if (!d_generator)
        setMaptype(NORMAL);

    d_generator->setNoCities(nocities);
}

void CreateScenario::setNoRuins(int noruins)
{
    debug("CreateScenario::setNoRuins")

    if (!d_generator)
        setMaptype(NORMAL);

    d_generator->setNoRuins(noruins);
}

void CreateScenario::setNoTemples(int notemples)
{
    debug("CreateScenario::setNoTemples")

    if (!d_generator)
        setMaptype(NORMAL);

    d_generator->setNoTemples(notemples);
}

void CreateScenario::setSize(int size)
{
    debug("CreateScenario::setSize")

    if (size < 0)
    {
        std::cerr <<"CreateScenario:: wrong size given\n";
        return;
    }

    d_size = size;

    //IMPORTANT!!
    GameMap::setSize(size);
}

void CreateScenario::addPlayer(std::string name, std::string armyset,
                                SDL_Color color, Player::TYPE type)
{
    debug("CreateScenario::addPlayer")

    Player* p;

    switch (type)
    {
        case Player::HUMAN:
            p = new RealPlayer(name, armyset, color);
            break;
        case Player::AI_FAST:
            p = new AI_Fast(name, armyset, color);
            break;
        case Player::AI_DUMMY:
            p = new AI_Dummy(name, armyset, color);
            break;
        case Player::NEUTRAL:
            p = new AI_Dummy(name, armyset, color);
            p->setType(Player::NEUTRAL);
            Playerlist::getInstance()->setNeutral(p);
            break;
        default:
            std::cerr <<"CreateScenario: wrong player type\n";
    }

    Playerlist::getInstance()->push_back(p);
}

int CreateScenario::getNoPlayers()
{
    return Playerlist::getInstance()->size();
}

Player* CreateScenario::getPlayer(int number)
{
    debug("CreateScenario::getPlayer")

    if (number >= static_cast<int>(Playerlist::getInstance()->size()))
        return 0;

    Playerlist::iterator it;
    for (it = Playerlist::getInstance()->begin(); number > 0; number--)
        it++;

    return (*it);
}

CreateScenario::MapType CreateScenario::getMaptype()
{
    //there is only one maptype yet
    return NORMAL;
}

int CreateScenario::getNoCities()
{
    if (!d_generator)
        return -1;

    return d_generator->getNoCities();
}

int CreateScenario::getNoRuins()
{
    if (!d_generator)
        return -1;

    return d_generator->getNoRuins();
}

int CreateScenario::getNoTemples()
{
    if (!d_generator)
        return -1;

    return d_generator->getNoTemples();
}

std::string CreateScenario::getMapTiles()
{
    debug("CreateScenario::getMapTiles")
    return d_tilesname;
}

bool CreateScenario::create()
{
    debug("CreateScenario::create")

    PG_Rect rect;
    rect.x = (d_uncle->Width() / 2) - 150;
    rect.y = (d_uncle->Height() / 2) - 50;
    rect.w = 300;
    rect.h = 100;
    MapCreationDialog dialog(d_uncle, rect);
    dialog.Show();

    dialog.setProgress(0, "Creating map");
    if (!createMap())
        return false;

    dialog.setProgress(30, "Distributing players");
    if (!distributePlayers())
        return false;

    dialog.setProgress(50, "Setting up cities");
    if (!setupCities())
        return false;

    dialog.setProgress(70, "Setting up ruins");
    if (!setupRuins())
        return false;

    dialog.setProgress(90, "Setting up players");
    if (!setupPlayers())
        return false;

    dialog.Hide();

    return true;
}

bool CreateScenario::dump(std::string filename)
{
    debug("CreateScenario::dump")

    //this function basically imitates the save function of GameScenario.
    //So, if the savegame style changes, these changes have to be done here, too
    bool retval = true;

    XML_Helper* helper = new XML_Helper(filename, std::ios::out,
                                        Configuration::s_zipfiles);
    retval &= helper->begin(FREELORDS_SAVEGAME_VERSION);
    retval &= helper->openTag("freelords");

    retval &= helper->openTag("scenario");
    retval &= helper->saveData("name", "random scenario");
    retval &= helper->saveData("type", GameScenario::SINGLE_PLAYER);
    retval &= helper->saveData("turnmode", d_turnmode);
    retval &= helper->saveData("turn", 0);
    retval &= helper->closeTag();

    //now save the rest of the data
    retval &= fl_counter->save(helper);

    retval &= GameMap::getInstance()->save(helper);
    retval &= Playerlist::getInstance()->save(helper);
    retval &= Templelist::getInstance()->save(helper);
    retval &= Ruinlist::getInstance()->save(helper);
    retval &= Citylist::getInstance()->save(helper);

    retval &= QuestsManager::getInstance()->save(helper);

    retval &= helper->closeTag();
    
    helper->close();
    delete helper;

    return retval;
}

bool CreateScenario::createMap()
{
    debug("CreateScenario::createMap")

    int temp;
    char type;
    char* map;
    
    //have the generator make the map...
    d_generator->MakeMap(d_size);
    
    //...fill the terrain...
    GameMap::getInstance(d_tilesname)->fill(d_generator);

    //...and create cities, temples and ruins
    map = d_generator->getBuildings(temp);
    
    for (int j = 0; j < d_size; j++)
        for (int i = 0; i < d_size; i++)
        {
            type = map[i + j*d_size];
            PG_Point pos;
            pos.x = i;      pos.y = j;

            switch (type)
            {
                case CHARACTER_TEMPLE:
                    Templelist::getInstance()->push_back(Temple(pos));
                    break;
                case CHARACTER_RUIN:
                    Ruinlist::getInstance()->push_back(Ruin(pos));
                    break;
                case CHARACTER_CITY_MAJOR:
                    Citylist::getInstance()->push_back(City(pos));
            }
        }
    //the other details such as giving names are done later
    
    return true;
}

bool CreateScenario::distributePlayers()
{
    debug("CreateScenario::distributePlayers")

   //how many cities we skip after a player assignment
    int cityskip = (Citylist::getInstance()->size() - 1) / (Playerlist::getInstance()->size() - 1) - 1;

    //some variables
    int skipping = 0;
    Playerlist::iterator pit = Playerlist::getInstance()->begin();
    if ((*pit) == Playerlist::getInstance()->getNeutral())
        pit++;

    for (Citylist::iterator cit = Citylist::getInstance()->begin();
        cit != Citylist::getInstance()->end(); cit++)
    {
        skipping++;
        if ((skipping >= cityskip) && (pit != Playerlist::getInstance()->end()))
        {
            (*cit).setPlayer(*pit);
            pit++;
            skipping = 0;
        
            if ((*pit) == Playerlist::getInstance()->getNeutral())
                pit++;
        }
        else
            (*cit).setPlayer(Playerlist::getInstance()->getNeutral());
    }

    return true;
}

bool CreateScenario::setupCities()
{
    debug("CreateScenario::setupCities")

    if (!loadCitynames())
    {
        std::cerr <<"Couldn't load citynames. Aborting scenario creation.\n";
        return false;
    }

    for (Citylist::iterator it = Citylist::getInstance()->begin();
        it != Citylist::getInstance()->end(); it++)
    {
        //1. set a reasonable cityname
        int randno;
        
        randno = rand() % d_citynames.size();
        (*it).setName(d_citynames[randno]);
        //remove the used name
        d_citynames[randno] = d_citynames[d_citynames.size() - 1];
        d_citynames.pop_back();

        //2. distribute the income a bit (TBD)

        //3. set the city production
        (*it).setRandomArmytypes();
        (*it).produceLargestArmy();
        if ((*it).getPlayer() == Playerlist::getInstance()->getNeutral())
            (*it).setProduction(-1);
        else
        {
            (*it).setProduction(0);
            //raise the city defense level (some eyecandy)
            (*it).setDefenseLevel(2);
        }
    }

    return true;
}

bool CreateScenario::setupRuins()
{
    debug("CreateScenario::setupRuins")

    //The aim of this function is to put a strong stack as sentinel in all
    //ruins.
   Ruinlist::iterator it;
   Armyset* myset = Armysetlist::getInstance()->getArmyset("default");
                //this set always exists

    for (it = Ruinlist::getInstance()->begin();
        it != Ruinlist::getInstance()->end(); it++)
    {
        Stack* s;
        Army* a = 0;
        PG_Point pos;

        pos = (*it).getPos();
        
        //create a strong stack:
        s = new Stack(0, pos);

        //(i) insert one random army first (no owning player). The army is an
        //arbitrary army except ships and heroes (this is an infinite loop if
        //the armyset contains only heroes and ships, but with enough impetus
        //you can break anything). Heroes and ships are just too mighty and
        //unusual to put them in ruins
        while (!a || (a->getArmyBonus() & (ARMY_SHIP | ARMY_HERO)))
        {
            if (a)
                delete a;
            a = new Army(myset, rand() % myset->getSize(), (Player*)0);
        }
            
        s->push_back(a);
        a = 0;

        //(ii) with some chance, insert a second army
        if ((rand() % 3) == 0)
        {
            while (!a || (a->getArmyBonus() & (ARMY_SHIP | ARMY_HERO)))
            {
                if (a)
                    delete a;
                a = new Army(myset, rand() % myset->getSize(), (Player*)0);
            }
            s->push_back(a);
            a = 0;
        }

        //(iii) and with another chance, insert still another army
        if ((rand() % 5) == 0)
        {
            while (!a || (a->getArmyBonus() & (ARMY_SHIP | ARMY_HERO)))
            {
                if (a)
                    delete a;
                a = new Army(myset, rand() % myset->getSize(), (Player*)0);
            }
            s->push_back(a);
        }

        //now mark this stack as guard
        (*it).setOccupant(s);
    }

    return true;
}

bool CreateScenario::setupPlayers()
{
    debug("CreateScenario::setupPlayers")

    //Up to now, this function only assigns each player a hero.

    //A player may later have more than one city, so loop over the players
    //instead of the cities
    Playerlist::iterator pit;
    
    for (pit = Playerlist::getInstance()->begin(); pit != Playerlist::getInstance()->end(); pit++)
    {
        if ((*pit) == Playerlist::getInstance()->getNeutral())
            continue;

        //find the first city of the player
        City* city = Citylist::getInstance()->getFirstCity(*pit);

        //put a hero in the city
        Hero* newhero = new Hero((*pit)->getArmyset(), rand() % 2, "", (*pit));
        city->addHero(newhero);
    }

    return true;
}

bool CreateScenario::loadCitynames()
{
    debug("CreateScenario::loadCitynames")

    std::ifstream namefile(File::getMiscFile("citynames").c_str());
    
    if (!namefile)
    {
        std::cerr <<"Critical Error: Couldn't open citynames data file\n";
        return false;
    }

    int counter;
    char buffer[101];
    buffer[100] = '\0';

    namefile >> counter;
    d_citynames.resize(counter);

    //with getline, the first call will get the first line to the end, i.e.
    //a newline character. Thus, we throw away the result of the first call.
    namefile.getline(buffer, 100);

    for (counter--; counter >= 0; counter--)
    {
        namefile.getline(buffer, 100);
        d_citynames[counter] = std::string(buffer);
    }

    return true;
}
