//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Library General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include "GraphicsCache.h"
#include "armysetlist.h"
#include "armyset_army.h"
#include "playerlist.h"
#include "Configuration.h"
#include "File.h"
#include "GameMap.h"

//#define debug(x) {cerr<<__FILE__<<": "<<__LINE__<<": "<<x<<endl;}
#define debug(x)

//some structures first

//this structure is the base class for storing cached objects. It stores the
//(army-, but can be extended) set, the type of army and the player which the
//surface is designed for (and the surface itself, of course).
struct ArmyCacheItem
{
    std::string graphset;
    Uint32 ref;
    Player* player;
    SDL_Surface* surface;
    int level;
};

//the structure to store buildings in
struct CityCacheItem
{
    int type;
    Player* player;
    SDL_Surface* surface;
};

//-----------------------------------------------------

GraphicsCache* GraphicsCache::s_instance = 0;
SDL_Surface* GraphicsCache::s_levelmask = 0;


GraphicsCache* GraphicsCache::getInstance()
{
    if (!s_instance)
        s_instance = new GraphicsCache();

    return s_instance;
}

void GraphicsCache::deleteInstance()
{
    if (!s_instance)
        return;

    delete s_instance;
    s_instance =0;
}

void GraphicsCache::clearInstance()
{
    if (!s_instance)
        return;

    s_instance->clear();
}

Uint32 GraphicsCache::getCacheSize()
{
    return Configuration::s_cacheSize;
}

GraphicsCache::GraphicsCache()
    :d_armysize(0), d_citysize(0)
{
    for (int i = 0; i < 3; i++)
    {
        d_citypic[i] = 0;
        d_citymask[i] = 0;
    }

    s_levelmask = File::getMiscPicture("level_mask.png");
}

GraphicsCache::~GraphicsCache()
{
    clear();

    for (int i = 0; i < 3; i++)
    {
        if (d_citypic[i])
            SDL_FreeSurface(d_citypic[i]);

        if (d_citymask[i])
            SDL_FreeSurface(d_citymask[i]);
    }
}

SDL_Surface* GraphicsCache::getArmyPic(std::string armyset, Uint32 army,
                                        Player* p, int level)
{
    debug("getting army pic" <<armyset <<" " <<army <<" " <<p->getName())

    std::list<ArmyCacheItem*>::iterator it;
    ArmyCacheItem* myitem;

    //special situation: ruin keepers don't belong to any player
    if (!p)
        p = Playerlist::getInstance()->getNeutral();
    
    for (it =d_armylist.begin(); it != d_armylist.end(); it++)
    {
        if (((*it)->graphset == armyset) && ((*it)->ref == army)
            && ((*it)->player ==p) && ((*it)->level==level))
        {
            myitem = (*it);
            
            //put the item on the first place (==last touched)
            d_armylist.erase(it);
            d_armylist.push_front(myitem);
            
            return myitem->surface;
        }
    }

    //We are still here, so the graphic is not in the cache. addArmyPic calls
    //checkPictures on its own, so we can simply return the surface
    myitem = addArmyPic(armyset, army, p, level);

    return myitem->surface;
}

SDL_Surface* GraphicsCache::getCityPic(int type, Player* p)
{
    debug("GraphicsCache::getCityPic " <<type <<", player" <<p->getName())
    
    if (!p)
    {
        std::cerr <<"GraphicsCache: wanted to get city pic without";
        std::cerr <<"player. Exiting...\n";
        exit(-1);
    }

    std::list<CityCacheItem*>::iterator it;
    CityCacheItem* myitem;

    for (it = d_citylist.begin(); it != d_citylist.end(); it++)
    {
        if (((*it)->type == type) && ((*it)->player == p))
        {
            myitem = (*it);

            //put the item in first place (last touched)
            d_citylist.erase(it);
            d_citylist.push_front(myitem);

            return myitem->surface;
        }
    }

    //no item found -> create a new one
    myitem = addCityPic(type, p);

    return myitem->surface;
}

SDL_Surface* GraphicsCache::getBuildingPic(int which, SDL_Surface** building,
                                            SDL_Surface** bmask, Player* p)
{
    SDL_Surface* mysurf = SDL_DisplayFormatAlpha(building[which]);

    //now we do basically the same as in addArmyPic, just with other
    //pointers
    SDL_Surface* mask = SDL_CreateRGBSurface(SDL_SWSURFACE, 100, 100,
                            bmask[which]->format->BitsPerPixel,0,0,0,0);

    //Set some color palette first!! Else we end up with a black surface.
    SDL_SetColors(mask, bmask[which]->format->palette->colors, 0,
                    bmask[which]->format->palette->ncolors);
                            
    SDL_BlitSurface(bmask[which], 0, mask, 0);

    //now change mask's palette to the player's one
    SDL_SetColors(mask, p->getPalette(), 0, 16);

    //set the first pixel as alpha value
    SDL_SetColorKey(mask, SDL_SRCCOLORKEY, 0);

    //blit mask over the army pic
    SDL_BlitSurface(mask, 0, mysurf, 0);

    //free the temporary surface
    SDL_FreeSurface(mask);

    return mysurf;
}

void GraphicsCache::checkPictures()
{
    if (d_armysize + d_citysize <= Configuration::s_cacheSize)
        //nothing to do
        return;
    
    /* Now I apply two kinds of checks:
     * We can easily calculate that an army pic has a maximum size of
     * 6400 Bytes, a city pic of 40000 Bytes (32 Bit color depth). We
     * guarantee that one pic always stays in the queue, since this
     * is unavoidable and the working of the rest of the class depends
     * on it.
     * If we are low on cachesize, we try to guarantee ~200kB (>30 pics)
     * for the army cache since this is the most important (the value
     * may be reached in large battles). Else, we distribute the cache
     * over armies and cities in a 3:1 ratio in favour of the armies.
     */

    if (Configuration::s_cacheSize < (1 << 18)) //low cache, 256kB
    {
        //kill cities until we reach the limit again
        while (d_citylist.size() > 1)
            eraseLastCityItem();
        
        if (d_armysize + d_citysize <= Configuration::s_cacheSize)
            return;

        //still not enough =>start to remove army pics
        while ((d_armysize + d_citysize > Configuration::s_cacheSize)
              && (d_armylist.size() > 1))
        {
            eraseLastArmyItem();
        }

        return;
    }
    
    //we are _not_ low on cache, so distribute the cache sizes a bit    
    while (d_armysize + d_citysize > Configuration::s_cacheSize)
    {
        if (d_armysize > 3 * d_citysize)
            eraseLastArmyItem();
        else
            eraseLastCityItem();
    }
}

ArmyCacheItem* GraphicsCache::addArmyPic(std::string armyset, Uint32 army,
                                                Player* p, int level)
{
    debug("loading army pic")

    ArmyCacheItem* myitem = new ArmyCacheItem();
    myitem->graphset = armyset;
    myitem->ref = army;
    myitem->player = p;
    myitem->level = level;

    //Now the most important part: load the army picture
    //First, copy the army picture and change it to the display format
    Armyset_army* basearmy = &(Armysetlist::getInstance()->getArmyset(armyset)->army[army]);
    myitem->surface = SDL_DisplayFormatAlpha(basearmy->getPixmap());

    //Now, if we have a mask ready, copy the mask over the picture. The mask
    //is <= 256 colors, we use the player's palette to colorize the mask. If
    //the armyset has no mask (happens during the transition phase), ignore this
    //step.
    if (basearmy->getMask())
    {
        SDL_Surface* mask = SDL_CreateRGBSurface(SDL_SWSURFACE, 40, 40,
                            basearmy->getMask()->format->BitsPerPixel,0,0,0,0);

        //Set some color palette first!! Else we end up with a black surface.
        SDL_SetColors(mask, basearmy->getMask()->format->palette->colors, 0,
                            basearmy->getMask()->format->palette->ncolors);
                            
        SDL_BlitSurface(basearmy->getMask(), 0, mask, 0);

        //now change tmp's palette to the player's one
        SDL_SetColors(mask, p->getPalette(), 0, 16);

        //set the first pixel as alpha value
        SDL_SetColorKey(mask, SDL_SRCCOLORKEY, 0);

        //blit mask over the army pic
        SDL_BlitSurface(mask, 0, myitem->surface, 0);

        //free the temporary surface
        SDL_FreeSurface(mask);
    }
    
    if (level>1)
      if (s_levelmask)
	{

	  SDL_Surface* mask = SDL_CreateRGBSurface(SDL_SWSURFACE, 40, 40,
						  s_levelmask->format->BitsPerPixel,0,0,0,0);

	  // a little hack while waiting for a complete level picture
	  if (level<3)
	    SDL_BlitSurface(s_levelmask, PG_Rect(40*(level-1), 0, 40, 40).SDLRect(), mask, 0);
	  else SDL_BlitSurface(s_levelmask, PG_Rect(80, 0, 40, 40).SDLRect(), mask, 0);
	  
	  //set the first pixel as alpha value
	  SDL_SetColorKey(mask, SDL_SRCCOLORKEY, 0);

	  //blit mask over the army pic
	  SDL_BlitSurface(mask,0, myitem->surface, 0);

	  //free the temporary surface
	  SDL_FreeSurface(mask);
	}

    //now the final preparation steps:
    //a) add the size
    int size = myitem->surface->w * myitem->surface->h;
    d_armysize += myitem->surface->format->BytesPerPixel * size;

    //b) add the entry to the list
    d_armylist.push_front(myitem);

    //c) check if the size is too large
    checkPictures();

    //we are finished, so return the pic
    return myitem;
}

CityCacheItem* GraphicsCache::addCityPic(int type, Player* p)
{
    //On the first run, load the city and mask picture.
    //We assume that either all or no surface are/is filled
    if ((d_citypic[0] == 0) || (d_citymask[0] == 0))
        loadCityPics();

    
    SDL_Surface* mysurf = SDL_DisplayFormatAlpha(d_citypic[type]);

    //now we do basically the same as in addArmyPic, just with other
    //pointers
    SDL_Surface* mask = SDL_CreateRGBSurface(SDL_SWSURFACE, 100, 100,
                            d_citymask[type]->format->BitsPerPixel,0,0,0,0);

    //Set some color palette first!! Else we end up with a black surface.
    SDL_SetColors(mask, d_citymask[type]->format->palette->colors, 0,
                  d_citymask[type]->format->palette->ncolors);
                            
    SDL_BlitSurface(d_citymask[type], 0, mask, 0);

    //now change mask's palette to the player's one
    SDL_SetColors(mask, p->getPalette(), 0, 16);

    //set the first pixel as alpha value
    SDL_SetColorKey(mask, SDL_SRCCOLORKEY, 0);

    //blit mask over the army pic
    SDL_BlitSurface(mask, 0, mysurf, 0);

    //free the temporary surface
    SDL_FreeSurface(mask);

    //now create the cache item and add the size
    CityCacheItem* myitem = new CityCacheItem();
    myitem->player = p;
    myitem->type = type;
    myitem->surface = mysurf;

    d_citylist.push_front(myitem);

    //add the size
    int size = mysurf->w + mysurf->h;
    d_citysize += size * mysurf->format->BytesPerPixel;

    //and check the size of the cache
    checkPictures();

    return myitem;
}

void GraphicsCache::clear()
{
    while (!d_armylist.empty())
        eraseLastArmyItem();

    while (!d_citylist.empty())
        eraseLastCityItem();
}

void GraphicsCache::eraseLastArmyItem()
{
    if (d_armylist.empty())
        return;

    //As the name suggests, this function erases the last item in the list.
    //Whenever an item is requested, it moves to the first position, so the
    //last item is the oldest and therefore propably most useless in the list.
    ArmyCacheItem* myitem = *(d_armylist.rbegin());
    d_armylist.erase(d_armylist.end()--);

    //don't forget to subtract the size from the size entry
    int size = myitem->surface->w * myitem->surface->h;
    d_armysize -= myitem->surface->format->BytesPerPixel * size;

    SDL_FreeSurface(myitem->surface);
    delete myitem;
}

void GraphicsCache::eraseLastCityItem()
{
    if (d_citylist.empty())
        return;

    CityCacheItem* myitem = *(d_citylist.rbegin());
    d_citylist.erase(d_citylist.end()--);

    int size = myitem->surface->w * myitem->surface->h;
    d_citysize -= myitem->surface->format->BytesPerPixel * size;

    SDL_FreeSurface(myitem->surface);
    delete myitem;
}

void GraphicsCache::loadCityPics()
{
    //GameMap has the actual tileset stored
    std::string tileset = GameMap::getInstance()->getTileSet()->getName();
    
    //load the city pictures
    SDL_Surface* citypics = File::getMapsetPicture(tileset, "misc/castles.png");
    SDL_Surface* citymasks = File::getMapsetMask(tileset, "misc/castles_mask.png");

    //copy alpha values, don't use them
    SDL_SetAlpha(citypics, 0, 0);
    
    if (citymasks->format->BitsPerPixel > 8)
    {
        std::cout <<"city masks have too much color depth, Exiting.\n";
        exit(-1);
    }
    
    for (int i = 0; i < 3; i++)
    {
        //copy the city image...
        SDL_Surface* tmp;
        SDL_PixelFormat* fmt = citypics->format;
        
        tmp = SDL_CreateRGBSurface(SDL_HWSURFACE, 100, 100, fmt->BitsPerPixel,
                            fmt->Rmask, fmt->Gmask, fmt->Bmask, fmt->Amask);

        SDL_BlitSurface(citypics, PG_Rect(i*100, 0, 100, 100).SDLRect(),
                        tmp, 0);

        d_citypic[i] = SDL_DisplayFormatAlpha(tmp);

        SDL_FreeSurface(tmp);
        
        //...and copy the mask image
        d_citymask[i] = SDL_CreateRGBSurface(SDL_HWSURFACE, 100, 100,
                        citymasks->format->BitsPerPixel,0,0,0,0);

        SDL_SetColors(d_citymask[i], citymasks->format->palette->colors, 0,
                        citymasks->format->palette->ncolors);

        SDL_BlitSurface(citymasks, PG_Rect(i*100, 0, 100, 100).SDLRect(),
                        d_citymask[i], 0);
    }

    SDL_FreeSurface(citypics);
    SDL_FreeSurface(citymasks);
}
