//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Library General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include <stdlib.h>
#include <time.h>
#include <iostream>
#include <fstream>
#include "MapGenerator.h"
#include "File.h"
#include "defs.h"

//#define debug(x) {std::cerr<<__FILE__<<": "<<__LINE__<<": "<<x<<std::endl;}
#define debug(x)


using namespace std;


MapGenerator::MapGenerator()
    //set reasonable default values
    :d_terrain(0), d_building(0), d_pswamp(2), d_pwater(7), d_pforest(3),
    d_phills(5), d_pmountains(5), d_nocities(20), d_notemples(25), d_noruins(25)
{
	//load file for city names
	ifstream namefile(File::getMiscFile("citynames").c_str());

	int counter;
	namefile >> counter;
	d_namelist.resize(counter);

	//with getline, the first call will return the first line to the end
	//of the line, i.e. a newline character. To prevent this, we do a throw
	//away call first
	char buffer[100];
	namefile.getline(buffer, 100);
	
	for(counter--; counter >= 0; counter--)
	{
		namefile.getline(buffer, 100);

		d_namelist[counter] = std::string(buffer);
	}
}

MapGenerator::~MapGenerator()
{
    if (d_terrain)
        delete[] d_terrain;
    if (d_building)
        delete[] d_building;
}

int MapGenerator::setNoCities(int nocities)
{
    if (nocities <= 0)
        return -1;

    int tmp = d_nocities;
    d_nocities = nocities;
    return tmp;
}

int MapGenerator::setNoRuins(int noruins)
{
    if (noruins < 0)
        return -1;

    int tmp = d_noruins;
    d_noruins = noruins;
    return tmp;
}

int MapGenerator::setNoTemples(int notemples)
{
    if (notemples < 0)
        return -1;

    int tmp = d_notemples;
    d_notemples = notemples;
    return tmp;
}

void MapGenerator::setPercentages(int pwater, int pforest, int pswamp,
                                    int phills, int pmountains)
{
    if ((pswamp < 0) || (pwater < 0) || (pforest < 0) || (phills < 0)
        || (pmountains < 0))
        return;

    if (pswamp + pwater + pforest + phills + pmountains > 100)
        return;

    d_pswamp = pswamp;
    d_pwater = pwater;
    d_pforest = pforest;
    d_phills = phills;
    d_pmountains = pmountains;
}

/** 
 * Generates a random map . The map is stored as a char array of size 
 * 100x100. Each character stands for something. We use :
 * M = mountains
 * h = hills
 * ~ = water
 * $ = forest
 * . = plains
 * _ = swamps
 * C = city/castle
 * r = ruins
 * T = temple
 *   = nothing
 * c = part of city/castle
 */
void MapGenerator::MakeMap(int size)
{
    d_size = size;

    //initialize terrain and building arrays
    d_terrain = new char[size*size];
    d_building = new char[size*size];
    for (int i = 0; i < size; i++)
        for (int j = 0; j < size; j++)
            d_building[i*size + j] = ' ';


    cout<<"Making random map"<<endl;
   
    cout<<"Flatening Plains  ... 10%"<<endl;
    makePlains();
    cout<<"Raising Hills     ... 20%"<<endl;
    makeTerrain(CHARACTER_HILLS,d_phills);
    cout<<"Raining Water     ... 30%"<<endl;
    makeTerrain(CHARACTER_WATER,d_pwater);
    cout<<"Planting Forest   ... 40%"<<endl;
    makeTerrain(CHARACTER_FOREST,d_pforest);
    cout<<"Watering Swamps   ... 50%"<<endl;
    makeTerrain(CHARACTER_SWAMP,d_pswamp);
    cout<<"Normalizing       ... 60%"<<endl;
    normalize();

    cout<<"Building Cities   ... 70%"<<endl;
    makeCities(d_nocities);
    cout<<"Ruining Ruins     ... 80%"<<endl;
    makeBuildings(CHARACTER_RUIN,d_noruins);
    cout<<"Spawning temples  ... 90%"<<endl;
    makeBuildings(CHARACTER_TEMPLE,d_notemples);
    cout<<"Done making map   ... 100%"<<endl;
}

char* MapGenerator::getMap(int& size)
{
    size = d_size;

    return d_terrain;
}

char* MapGenerator::getBuildings(int& size)
{
    size = d_size;

    return d_building;
}

std::vector<std::string>* MapGenerator::getCitynames()
{
    return &d_citynames;
}

/* Nothing much to do here, just fill up everything with '.' 
 * as we are making the default canvas for the map.
 */
void MapGenerator::makePlains()
{
    for(int j = 0; j < d_size; j++)
        for(int i = 0; i < d_size; i++)
            d_terrain[j*d_size + i]=CHARACTER_GRASS;
}


/**
 * Makes Terrains.
 * The algo is as follows :
 * 1. Find a random starting location
 * 2. chose a random direction , if x is the starting location, the direction
 *    can be from 0-7 as follows :
 *    +-+-+-+
 *    |0|1|2|
 *    +-+-+-+
 *    |7|x|3|
 *    +-+-+-+
 *    |6|5|4|
 *    +-+-+-+
 * 3. Check if there is some other terrain there ('.' = plains is okay)
 * 4. Reserve the location, and chose another direction
 */
void MapGenerator::makeTerrain(char c,int percent)
{
    // calculate the total number of tiles for this terrain
    int terrain = d_size*d_size*percent / 100;
    int xdir[]= {-1, 0, 1, 1, 1, 0,-1,-1};
    int ydir[]= { 1, 1, 1, 0,-1,-1,-1, 0};

    int count = 0;  // counter 
    int done = 0;
    int loopcount = 0;
    int tmpx = 0;
    int tmpy = 0;
    int direction = 0;
    int x,y;    // locations
    while(count != terrain)
    {
        // find a random starting location
        do{
            x = rand()%d_size;
            y = rand()%d_size;
        }while( d_terrain[y*d_size + x]!=CHARACTER_GRASS);
        // starting position
        d_terrain[y*d_size + x]=c;
        count++;
        while(count != terrain)
        {
            //find direction
            done = 0,loopcount = 0;
            direction  = rand()%8;
            while( !done )
            {
                direction +=loopcount;
                // NOTE: potential hazard here as if the direction
                // becomes more than the max value, it will always be
                // set to 0 (north west).
                if(direction>7)
                    direction = 0;
                tmpx = x+xdir[direction];
                tmpy = y+ydir[direction];
                if(tmpx < d_size && tmpx >= 0 && tmpy < d_size && tmpy >= 0
                    && d_terrain[tmpy*d_size + tmpx]==CHARACTER_GRASS)
                    done =1;
                if(++loopcount > 7)
                    done =1;
                

            }
            if(loopcount>=7 || tmpx <0 || tmpy <0 || tmpx > d_size || tmpy >d_size)
                break; // find another starting point
            x=tmpx;
            y=tmpy;
            d_terrain[y*d_size + x]=c;
            count++;
        }

    }// end while
    
    
}
void MapGenerator::makeCities(int cities)
{
    int x,y,city_count;
    int iterations = 0;
    
    // place the cities
    city_count = 0;

    while(city_count < cities)
    {
        x = rand()%(d_size-2);
        y = rand()%(d_size-2);
        if((d_terrain[y*d_size +x] == CHARACTER_WATER
            || d_terrain[y*d_size + x+1] == CHARACTER_WATER 
            || d_terrain[(y+1)*d_size + x] == CHARACTER_WATER
            || d_terrain[(y+1)*d_size + x+1] == CHARACTER_WATER)
            && (iterations < 10))
        {
            iterations++;
            continue;
        }
        
        // check if we can put the building
        if(!canPutBuilding(x,y) &&
            ((iterations < 10)
            || (d_terrain[y*d_size+x] != CHARACTER_WATER)))
        {
            iterations++;
            continue;
        }
        
        d_building[y*d_size + x]    =CHARACTER_CITY_MAJOR;
        d_building[y*d_size + x+1]  =CHARACTER_CITY_MINOR;
        d_building[(y+1)*d_size + x]  =CHARACTER_CITY_MINOR;
        d_building[(y+1)*d_size + x+1]=CHARACTER_CITY_MINOR;

        //cities shall only sit on grass tiles
        d_terrain[y*d_size + x]       = CHARACTER_GRASS;
        d_terrain[y*d_size + x+1]     = CHARACTER_GRASS;
        d_terrain[(y+1)*d_size + x]   = CHARACTER_GRASS;
        d_terrain[(y+1)*d_size + x+1] = CHARACTER_GRASS;

        city_count ++;
        iterations=0;
    }

    //give the cities names (order from top left to down right)
    d_citynames.resize(cities);

    if (d_citynames.size() > d_namelist.size())
    //mustn't happen that there are more cities than names for them
    {
    	std::cerr<<"Not enough names for naming cities on random map, exiting\n";
        std::cerr<<d_namelist.size() <<" names existed for " <<city_count <<" cities\n";
        exit(-1);
    }

    for (city_count = 0; city_count < cities; city_count++)
    {
    	int rand_no;
        rand_no = rand() % d_namelist.size();
        d_citynames[city_count] = d_namelist[rand_no];
        d_namelist[rand_no] = d_namelist[d_namelist.size()-1];
        d_namelist.pop_back();	//delete the already used name
    }
}

void MapGenerator::makeBuildings(char c, int building)
{
    int x,y;
    int i = 0;
    int iterations = 0;
    
    while (i < building)
    {        
        x = rand()%d_size;
        y = rand()%d_size;
        
        //If number of iterations is smaller 10, look for a suitable
        //place. If this number is exceeded, place the temple on an
        //island if neccessary. Try to write down the condition. It is
        //difficult because I have contracted it a bit.
        if (!canPutBuilding(x, y) && 
            ((iterations < 10)
            || (d_terrain[y*d_size+x] != CHARACTER_WATER)))
        {
            iterations++;
            continue;
        }

        if (iterations >= 10)
            d_terrain[y*d_size + x] = CHARACTER_GRASS;
        d_building[y*d_size + x] = c;
        
        iterations = 0;
        i++;
    }
}


/** 
 * canPutBuilding
 * Checks if we can put a building at the specified place. 
 * If we are on a sqare with water, we cannot put it.
 * If out of the surrounding squares, we have water in more than
 * 8 squares, we cannot.
 * int x = x position
 * int y = y position
 */

bool MapGenerator::canPutBuilding(int x,int y)
{
  //Was unused
  //int count=0;        // stores the count of water squares.
 
    // if the building is on water, return false
    if (d_terrain[y*d_size +x] == CHARACTER_WATER)
        return false;
        
    //if the building is close to the map boundaries, return false
    if ((x < 1) || (x > d_size-4) || (y < 1) || (y > d_size-4))
        return false;
        
    //if there is another building too close, return false
    for (int i = -3; i <= 3; i++)
        for (int j = -3; j <= 3; j++)
        {
            if ((y+j < 0) || (y+j >= d_size) || (x+i < 0) || (x+i >= d_size))
                continue;
            if (d_building[(y+j)*d_size + x+i] != ' ')
                return false;
        }
    
    // everything okay here ! return true
    return true;
    
}

/** 
 * normalize
 * Here we try to normalize the map. 
 * Currently only normalizes the hills and removes water puddles
 */
void MapGenerator::normalize()
{
    int count=0;
    char surrounding='\0';
    
    for(int j=1;j<(d_size-2);j++)
    {
        for(int i=1;i<(d_size-2);i++)
        {
            // If the hills is surrounded by other hills
            // in the NEWS directions, promote to mountains
            if(d_terrain[j*d_size +i] == CHARACTER_HILLS)
            {
                if ((d_terrain[(j-1)*d_size +i] == CHARACTER_HILLS
                     || d_terrain[(j-1)*d_size +i]== CHARACTER_MOUNTAINS )&&
                    (d_terrain[(j+1)*d_size +i] == CHARACTER_HILLS
                     || d_terrain[(j+1)*d_size +i]== CHARACTER_MOUNTAINS )&&
                    (d_terrain[j*d_size +i-1] == CHARACTER_HILLS
                     || d_terrain[j*d_size +i-1]== CHARACTER_MOUNTAINS )&&
                    (d_terrain[j*d_size +i+1] == CHARACTER_HILLS
                     || d_terrain[j*d_size +i+1]== CHARACTER_MOUNTAINS))
                {
                    // the probability of converting the hill to mountain
                    // is about 40%
                    if(rand()%100 < 40 )
                        d_terrain[j*d_size +i] = CHARACTER_MOUNTAINS;
                }
            } // end if hill
            
            //the following code is for the real normalization. It removes 
            //appendixes from a maptile type for the case that one or two tiles
            //of one type are embedded in another one (e.g. water puddles)
            count = 0;
            for(int x=i-1;x<=i+1;x++)
            {
                for(int y=j-1;y<=j+1;y++)
                {
                    if(d_terrain[y*d_size +x] == d_terrain[j*d_size +i])
                        count++;
                    else
                        surrounding = d_terrain[y*d_size +x];
                }
            }
            if(count < 2)
            {
                d_terrain[j*d_size +i]=surrounding;
            }
            if ((count == 2) && (rand() %2))
            {
                //give the tile the chance of survival
                d_terrain[j*d_size +i] = surrounding;
            }
        }
    }


    //now, with the last normalization, we may have created isolated tiles.
    //So go through all tiles again and remove isolated tiles.
    for (int j = 0; j < d_size; j++)
        for (int i = 0; i < d_size; i++)
        {
            count = 0;
            for(int x=i-1;x<=i+1;x++)
            {
                if ((x < 0) || (x >= d_size))
                {
                    count += 3; //we treat bordering tiles as if surrounded
                                //by the same type
                    continue;
                }
                
                for(int y=j-1;y<=j+1;y++)
                {
                    if ((y < 0) || (y >= d_size))
                    {
                        count++;
                        continue;
                    }
                    if(d_terrain[y*d_size +x] == d_terrain[j*d_size +i])
                        count++;
                    else
                        surrounding = d_terrain[y*d_size +x];
                }
            }
            if (count == 0)
                d_terrain[j*d_size +i] = surrounding;
        }
            
}

