//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Library General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include <iostream>
#include <pgdraw.h>
#include "smallmap.h"
#include "stacklist.h"
#include "citylist.h"
#include "ruinlist.h"
#include "templelist.h"
#include "city.h"
#include "ruin.h"
#include "temple.h"
#include "playerlist.h"
#include "player.h"
#include "GameMap.h"

using namespace std;

//#define debug(x) {cerr<<__FILE__<<": "<<__LINE__<<": "<<x<<endl;}
#define debug(x)

SmallMap::SmallMap(PG_Widget* parent, PG_Rect rect, int x_squares, int y_squares)
	:PG_Widget(parent, rect, true), d_pressed(false), d_r(50), d_g(50),
     d_b(50), d_add(true), d_timerID(0)
{
	d_viewrect = new PG_Rect(1, 1, x_squares, y_squares);

	SDL_PixelFormat* fmt = GetScreenSurface()->format;
	d_staticMap = SDL_CreateRGBSurface(SDL_HWSURFACE, GameMap::getSize(),
            GameMap::getSize(), fmt->BitsPerPixel, fmt->Rmask, fmt->Gmask,
			fmt->Bmask, fmt->Amask);
	
	// Draw map
	SDL_Color c;
	for(int i = 0; i < GameMap::getSize(); i++)
	{
		for(int j = 0; j < GameMap::getSize(); j++)
		{
			c = GameMap::getInstance()->getTile(i,j)->getColor();
			PG_Draw::SetPixel(i+1, j+1, c.r, c.g, c.b, d_staticMap);
		}
	}

    restartTimer();
}

SmallMap::~SmallMap()
{
	delete d_viewrect;
	SDL_FreeSurface(d_staticMap);
    interruptTimer();
}

void SmallMap::eventDraw(SDL_Surface* surface, const PG_Rect& rect)
{
    debug("eventDraw()");

    SDL_BlitSurface(d_staticMap, 0, surface, PG_Rect(rect.x, rect.y, rect.w, rect.h).SDLRect());
    DrawBorder(rect, 1);

    int xpos, ypos;
    int xsize, ysize;
    int mapsize = GameMap::getSize();

    // Draw all cities
    for (Citylist::iterator it = Citylist::getInstance()->begin();
        it != Citylist::getInstance()->end(); it++)
    {	
        SDL_Color c = (*it).getPlayer()->getColor();

        //We draw a rect of 4x4 pixels around an inner rect where the city is.
        //This can create some problems when the city is at the border of the
        //map. Then the outer rect would be off the map. So we try to fix this.
        xpos = (*it).getPos().x -1;
        ypos = (*it).getPos().y -1;
        xsize = ysize = 4;
        
        if (xpos < 0)
        {
            xpos++;
            xsize--;
        }
        if (xpos+xsize > mapsize)
            xsize--;
        if (ypos < 0)
        {
            ypos++;
            ysize--;
        }
        if (ypos+ysize > mapsize)
            ysize--;
        
        DrawRectWH(xpos, ypos, xsize, ysize, 0, 0, 0);
        //the inner rect is where the city is, so it should always be all right,
        //else we are in trouble
        DrawRectWH((*it).getPos().x, (*it).getPos().y, 2, 2, c.r, c.g, c.b);
    }

    // Draw all ruins
    for (Ruinlist::iterator it = Ruinlist::getInstance()->begin();
        it != Ruinlist::getInstance()->end(); it++)
    {
        //circumvent wrong memory access with ruins as well
        xpos = (*it).getPos().x;
        ypos = (*it).getPos().y;
        xsize = ysize = 2;

        if (xpos+xsize > mapsize)
            xsize--;
        if (ypos+ysize > mapsize)
            ysize--;
   
        // draw a yellow rect
        DrawRectWH(xpos, ypos, xsize, ysize, 255, 255, 0);	
    }

    // Draw temples
    for (Templelist::iterator it = Templelist::getInstance()->begin();
        it != Templelist::getInstance()->end(); it++)
    {
        //draw a magenta Pi-like symbol (3x3)
        xpos = (*it).getPos().x - 1;
        ypos = (*it).getPos().y - 1;
        xsize = 3;

        //draw the upper bar
        if (ypos >= 0)
        {
            if (xpos < 0)
            {
                xpos++;
                xsize--;
            }
            if (xpos+xsize > mapsize)
                xsize--;
            DrawHLine(xpos, ypos, xsize, 255, 0, 255);
        }

        //draw the left bar
        xpos = (*it).getPos().x -1;
        ypos = (*it).getPos().y;
        ysize = 2;
        if (xpos >= 0)
        {
            if (ypos+ysize > mapsize)
                ysize--;
            DrawVLine(xpos, ypos, ysize, 255, 0, 255);
        }

        //and draw the right bar
        xpos = (*it).getPos().x +1;
        //ypos and ysize can be taken from he previous initialization
        if (xpos < mapsize)
            DrawVLine(xpos, ypos, ysize, 255, 0, 255);
    }

    // Draw stacks
    for (Playerlist::iterator pit = Playerlist::getInstance()->begin();
        pit != Playerlist::getInstance()->end(); pit++)
    {
        Stacklist* mylist = (*pit)->getStacklist();
        SDL_Color c = (*pit)->getColor();
        for (Stacklist::iterator it= mylist->begin(); it != mylist->end(); it++)
        {
            //don't draw stacks in cities, they could hardly be identified
            Maptile* mytile = GameMap::getInstance()->getTile((*it)->getPos().x, (*it)->getPos().y);
            if (mytile->getBuilding() == Maptile::CITY)
                continue;

            //draw a cross for the stack
            //first the hoizontal bar
            xpos = (*it)->getPos().x -1;
            ypos = (*it)->getPos().y;
            xsize = 3;

            if (xpos < 0)
            {
                xpos++;
                xsize--;
            }
            if (xpos+xsize > mapsize)
                xsize--;
                
            DrawHLine(xpos, ypos, xsize, c.r, c.g, c.b);

            //and now the vertical line
            xpos = (*it)->getPos().x;
            ypos = (*it)->getPos().y -1;
            ysize = 3;

            if (ypos < 0)
            {
                ypos++;
                ysize--;
            }
            if (ypos+ysize > mapsize)
                ysize--;
                
            DrawVLine(xpos, ypos, ysize, c.r, c.g, c.b);
        }
    }

	drawViewrect();
}

// do some animation of the viewrect
void SmallMap::drawViewrect()
{
	if (d_r == 255) d_add = false;
	if (d_r == 160) d_add = true;
 
    if (d_add) {++d_r; ++d_g; ++d_b;}
    else {--d_r; --d_g; --d_b;} 
	
	// Draw the rectangle that shows the viewed part of the map
	DrawRectWH(d_viewrect->my_xpos, d_viewrect->my_ypos,
               d_viewrect->my_width, d_viewrect->my_height, d_r, d_g, d_b);
    SDL_UpdateRect(GetWidgetSurface(), d_viewrect->my_xpos, d_viewrect->my_ypos,
               d_viewrect->my_width, d_viewrect->my_height);
}

Uint32 SmallMap::eventTimer(PG_TimerID id, Uint32 interval)
{
    drawViewrect();
    Update();
    PG_TimerObject::eventTimer(id, interval);
	return interval;
}

// Pressed in the drawing area to set viewrect

bool SmallMap::eventMouseButtonDown(const SDL_MouseButtonEvent* event)
{
	if(event->button == SDL_BUTTON_LEFT)
	{
		d_pressed = true;
		setViewrect(event->x, event->y);
	}
	return true;
}

// Released in the drawing area to set viewrect

bool SmallMap::eventMouseButtonUp(const SDL_MouseButtonEvent* event)
{
	if(event->button == SDL_BUTTON_LEFT) d_pressed = false;
	return true;
}

// This eventhandler is called when moving the mouse with pressed button in
// the drawing area to move viewrect

bool SmallMap::eventMouseMotion(const SDL_MouseMotionEvent* event)
{
	if (d_pressed) setViewrect(event->x, event->y);
	return true;
}

PARAGUI_CALLBACK(SmallMap::changedViewrect)
{
	debug("changedViewrect()");
	SDL_Rect* rect = (SDL_Rect*)clientdata;
	d_viewrect->my_xpos = rect->x;
	d_viewrect->my_ypos = rect->y;
	Redraw();
	return true;
}

void SmallMap::stackMoved()
{
	debug("stackMoved()");
	Redraw();
}

void SmallMap::setViewrect(int x, int y)
{
    //the x/y coordinates are absolute to the whole programm, so subtract the
    //position of the widget and a bit more so the mouse isn't in the top left
    //corner
	x -= 5 + my_xpos;
	y -= 5 + my_ypos;

	if (x < 0) x = 0;
	if (y < 0) y = 0;
	if (x > (GameMap::getSize() - d_viewrect->my_width)) 
        x = GameMap::getSize() - d_viewrect->my_width;
	if (y > (GameMap::getSize() - d_viewrect->my_height))
        y = GameMap::getSize() - d_viewrect->my_height;

	d_viewrect->my_xpos = x;
	d_viewrect->my_ypos = y;

	Redraw();
    schangingViewrect.emit();
}

void SmallMap::interruptTimer()
{
    if (d_timerID != 0)
        RemoveTimer(d_timerID);

    d_timerID = 0;
}

void SmallMap::restartTimer()
{
    if (d_timerID == 0)
        d_timerID = AddTimer(TIMER_SMALLMAP_REFRESH);
}
